@@ -75,6 +75,8 @@ struct _WnckScreenPrivate
   Window xroot;
   Screen *xscreen;
 
+  int orig_event_mask;
+
   /* in map order */
   GList *mapped_windows;
   /* in stacking order */
@@ -513,6 +515,11 @@ wnck_screen_finalize (GObject *object)
 
   screen = WNCK_SCREEN (object);
 
+  _wnck_select_input (screen->priv->xscreen,
+                      screen->priv->xroot,
+                      screen->priv->orig_event_mask,
+                      FALSE);
+
   unqueue_update (screen);
 
   for (tmp = screen->priv->stacked_windows; tmp; tmp = tmp->next)
@@ -595,9 +602,10 @@ wnck_screen_construct (Display    *display,
 
   screen->priv->bg_pixmap = None;
 
-  _wnck_select_input (screen->priv->xscreen,
-                      screen->priv->xroot,
-                      PropertyChangeMask);
+  screen->priv->orig_event_mask = _wnck_select_input (screen->priv->xscreen,
+                                                      screen->priv->xroot,
+                                                      PropertyChangeMask,
+                                                      TRUE);
 
   screen->priv->need_update_workspace_list = TRUE;
   screen->priv->need_update_stack_list = TRUE;
@@ -2755,3 +2763,26 @@ _wnck_screen_change_workspace_name (WnckScreen *screen,
 
   g_free (names);
 }
+
+void
+_wnck_screen_shutdown_all (void)
+{
+  int i;
+  Display *display;
+
+  if (screens == NULL)
+    return;
+
+  display = _wnck_get_default_display ();
+
+  for (i = 0; i < ScreenCount (display); ++i)
+    {
+      if (screens[i] != NULL) {
+        g_object_unref (screens[i]);
+        screens[i] = NULL;
+      }
+    }
+
+  g_free (screens);
+  screens = NULL;
+}
