@@ -78,6 +78,7 @@ struct _WnckWindowPrivate
   WnckClassGroup *class_group;
   Window group_leader;
   Window transient_for;
+  int orig_event_mask;
   GdkRectangle icon_geometry;
   char *name;
   char *icon_name;
@@ -211,6 +212,16 @@ static WnckWindow* find_last_transient_for (GList *windows,
 
 static guint signals[LAST_SIGNAL] = { 0 };
 
+void
+_wnck_window_shutdown_all (void)
+{
+  if (window_hash != NULL)
+    {
+      g_hash_table_destroy (window_hash);
+      window_hash = NULL;
+    }
+}
+
 static void
 wnck_window_init (WnckWindow *window)
 {
@@ -222,6 +233,7 @@ wnck_window_init (WnckWindow *window)
   window->priv->class_group = NULL;
   window->priv->group_leader = None;
   window->priv->transient_for = None;
+  window->priv->orig_event_mask = 0;
   window->priv->icon_geometry.width = -1; /* invalid cached value */
   window->priv->name = NULL;
   window->priv->icon_name = NULL;
@@ -416,6 +428,11 @@ wnck_window_finalize (GObject *object)
 
   window = WNCK_WINDOW (object);
 
+  _wnck_select_input (WNCK_SCREEN_XSCREEN (window->priv->screen),
+                      window->priv->xwindow,
+                      window->priv->orig_event_mask,
+                      FALSE);
+
   unqueue_update (window);
 
   if (window->priv->app)
@@ -453,6 +470,8 @@ wnck_window_finalize (GObject *object)
   g_free (window->priv->res_name);
   window->priv->res_name = NULL;
 
+  window->priv->xwindow = None;
+
   G_OBJECT_CLASS (wnck_window_parent_class)->finalize (object);
 }
 
@@ -502,7 +521,8 @@ _wnck_window_create (Window      xwindow,
   Screen     *xscreen;
 
   if (window_hash == NULL)
-    window_hash = g_hash_table_new (_wnck_xid_hash, _wnck_xid_equal);
+    window_hash = g_hash_table_new_full (_wnck_xid_hash, _wnck_xid_equal,
+                                         NULL, g_object_unref);
 
   g_return_val_if_fail (g_hash_table_lookup (window_hash, &xwindow) == NULL,
                         NULL);
@@ -521,9 +541,10 @@ _wnck_window_create (Window      xwindow,
    * that's why we select the union of the mask we want for Application
    * and the one we want for window
    */
-  _wnck_select_input (xscreen,
-                      window->priv->xwindow,
-                      WNCK_APP_WINDOW_EVENT_MASK);
+  window->priv->orig_event_mask =_wnck_select_input (xscreen,
+                                                     window->priv->xwindow,
+                                                     WNCK_APP_WINDOW_EVENT_MASK,
+                                                     TRUE);
 
   /* Default the group leader to the window itself; it is set in
    * update_wmhints() if a different group leader is specified.
@@ -571,16 +592,17 @@ _wnck_window_create (Window      xwindow,
 void
 _wnck_window_destroy (WnckWindow *window)
 {
-  g_return_if_fail (wnck_window_get (window->priv->xwindow) == window);
+  Window xwindow = window->priv->xwindow;
 
-  g_hash_table_remove (window_hash, &window->priv->xwindow);
+  g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  g_return_if_fail (wnck_window_get (window->priv->xwindow) == NULL);
+  g_return_if_fail (wnck_window_get (xwindow) == window);
 
-  window->priv->xwindow = None;
+  g_hash_table_remove (window_hash, &xwindow);
 
-  /* remove hash's ref on the window */
-  g_object_unref (G_OBJECT (window));
+  /* Removing from hash also removes the only ref WnckWindow had */
+
+  g_return_if_fail (wnck_window_get (xwindow) == NULL);
 }
 
 static Display *
